
/*
 * opengl programming framework
 *
 * based on original from Jeff Molofee's tutorial #4
 */

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <complex>
using Complex = std::complex<float>;

// key input, true if key is down. Use windows' VK_* indexes or 'C'
static bool keys[256];

// global state
static float rtri;
static float rquad;

void initialize ()
{
	glShadeModel (GL_SMOOTH);
	glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth (1.0f);
	glEnable (GL_DEPTH_TEST);
	glDepthFunc (GL_LEQUAL);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

bool update (float deltaT)
{
	if(keys[VK_SPACE]) rtri += 360*deltaT;
	rquad -= 15*deltaT;

	return true;
}

void render ()
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef (-1.5f, 0.0f, -6.0f);
	glRotatef (rtri, 0.0f, 1.0f, 0.0f);
	glBegin (GL_TRIANGLES);
		glColor3f (1.0f, 0.0f, 0.0f);
		glVertex3f (0.0f, 1.0f, 0.0f);
		glColor3f (0.0f, 1.0f, 0.0f);
		glVertex3f (-1.0f, -1.0f, 0.0f);
		glColor3f (0.0f, 0.0f, 1.0f);
		glVertex3f (1.0f, -1.0f, 0.0f);
	glEnd();

	glLoadIdentity();
	glTranslatef (1.5f, 0.0f, -6.0f);
	glRotatef (rquad, 1.0f, 0.0f, 0.0f);
	glColor3f (0.5f, 0.5f, 1.0f);
	glBegin (GL_QUADS);
		glVertex3f (-1.0f, 1.0f, 0.0f);
		glVertex3f (1.0f, 1.0f, 0.0f);
		glVertex3f (1.0f, -1.0f, 0.0f);
		glVertex3f (-1.0f, -1.0f, 0.0f);
	glEnd();
}

// everytime the window gets resized, this function is called with a new size
void handle_resize (int width, int height)
{
	if (height == 0) {
		height = 1;
	}

	glViewport (0, 0, width, height);

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();

	gluPerspective (45.0f, (float) width / (float) height, 0.1f, 100.0f);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity();
}


/*
 * WINDOWS-SPECIFIC PART
 */

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "gdi32.lib")

static HDC hDC = NULL;
static HGLRC hRC = NULL;
static HWND hWnd = NULL;
static HINSTANCE hInstance;
static bool active = true;

LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);

void KillGLWindow ()
{
	if (hRC) {
		if (!wglMakeCurrent (NULL, NULL)) {
			MessageBox (NULL, "Release Of DC And RC Failed.", "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION);
		}

		if (!wglDeleteContext (hRC)) {
			MessageBox (NULL, "Release Rendering Context Failed.", "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION);
		}
		hRC = NULL;
	}

	if (hDC && !ReleaseDC (hWnd, hDC)) {
		MessageBox (NULL, "Release Device Context Failed.", "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION);
		hDC = NULL;
	}

	if (hWnd && !DestroyWindow (hWnd)) {
		MessageBox (NULL, "Could Not Release hWnd.", "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION);
		hWnd = NULL;
	}

	if (!UnregisterClass ("OpenGL", hInstance)) {
		MessageBox (NULL, "Could Not Unregister Class.", "SHUTDOWN ERROR", MB_OK | MB_ICONINFORMATION);
		hInstance = NULL;
	}
}

bool CreateGLWindow (const char* title, int width, int height, unsigned char bits, unsigned char depth)
{
	GLuint PixelFormat;
	WNDCLASS wc;
	DWORD dwExStyle;
	DWORD dwStyle;
	RECT WindowRect;
	WindowRect.left = (long) 0;
	WindowRect.right = (long) width;
	WindowRect.top = (long) 0;
	WindowRect.bottom = (long) height;

	hInstance = GetModuleHandle (NULL);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC) WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon (NULL, IDI_WINLOGO);
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "OpenGL";

	if (!RegisterClass (&wc)) {
		MessageBox (NULL, "Failed To Register The Window Class.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	dwStyle = WS_OVERLAPPEDWINDOW;

	AdjustWindowRectEx (&WindowRect, dwStyle, false, dwExStyle);


	if (! (hWnd = CreateWindowEx (dwExStyle,
	                              "OpenGL",
	                              title,
	                              dwStyle |
	                              WS_CLIPSIBLINGS |
	                              WS_CLIPCHILDREN,
	                              0, 0,
	                              WindowRect.right - WindowRect.left,
	                              WindowRect.bottom - WindowRect.top,
	                              NULL,
	                              NULL,
	                              hInstance,
	                              NULL))) {
		KillGLWindow();
		MessageBox (NULL, "Window Creation Error.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof (PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		bits,
		0, 0, 0, 0, 0, 0,
		0,
		0,
		0,
		0, 0, 0, 0,
		depth,
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	if (! (hDC = GetDC (hWnd))) {
		KillGLWindow();
		MessageBox (NULL, "Can't Create A GL Device Context.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	if (! (PixelFormat = ChoosePixelFormat (hDC, &pfd))) {
		KillGLWindow();
		MessageBox (NULL, "Can't Find A Suitable PixelFormat.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	if (!SetPixelFormat (hDC, PixelFormat, &pfd)) {
		KillGLWindow();
		MessageBox (NULL, "Can't Set The PixelFormat.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	if (! (hRC = wglCreateContext (hDC))) {
		KillGLWindow();
		MessageBox (NULL, "Can't Create A GL Rendering Context.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	if (!wglMakeCurrent (hDC, hRC)) {
		KillGLWindow();
		MessageBox (NULL, "Can't Activate The GL Rendering Context.", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	ShowWindow (hWnd, SW_SHOW);
	SetForegroundWindow (hWnd);
	SetFocus (hWnd);
	handle_resize (width, height);

	return true;
}

LRESULT CALLBACK WndProc (HWND hWnd,
                          UINT uMsg,
                          WPARAM wParam,
                          LPARAM lParam)
{
	switch (uMsg) {
	case WM_ACTIVATE: {
		if (!HIWORD (wParam)) {
			active = true;
		} else {
			active = false;
		}

		return 0;
	}

	case WM_SYSCOMMAND: {
		switch (wParam) {
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:
			return 0;
		}
		break;
	}

	case WM_CLOSE: {
		PostQuitMessage (0);
		return 0;
	}

	case WM_KEYDOWN: {
		keys[wParam] = true;
		return 0;
	}

	case WM_KEYUP: {
		keys[wParam] = false;
		return 0;
	}

	case WM_SIZE: {
		handle_resize (LOWORD (lParam), HIWORD (lParam));
		return 0;
	}
	}


	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}

int WINAPI WinMain (HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine,
                    int nCmdShow)
{
	bool done = false;

	LARGE_INTEGER timer_freq, timer_last;
	QueryPerformanceFrequency(&timer_freq);

	if (!CreateGLWindow ("OpenGL", 800, 600, 32, 16)) {
		return 0;
	}

	initialize();

	QueryPerformanceCounter(&timer_last);
	while (!done) {
		MSG msg;
		if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) {
				done = true;
			} else {
				TranslateMessage (&msg);
				DispatchMessage (&msg);
			}
			continue;
		}

		LARGE_INTEGER timer_now;
		QueryPerformanceCounter(&timer_now);
		float timeDiff=(timer_now.QuadPart-timer_last.QuadPart)/(float)timer_freq.QuadPart;
		timer_last=timer_now;
		if (!update(timeDiff)) done = true;
		
		if (active) render();
		if (keys[VK_ESCAPE]) done = true;
		SwapBuffers (hDC);
	}


	KillGLWindow();
	return 0;
}
